-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local merchantSearch;
local merchantItemMasks = {};

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to enable/disable Merchant Search
--
function NigglesMerchantSearchEnable()
  -- Local Variables
  local frame;
  local texture;

  -- Check if the edit box and textures need to be created
  if ((NigglesDBC.settings.UiMerchantSearch) and
      (merchantSearch == nil)) then
    -- Create the edit box
    merchantSearch = CreateFrame("EDITBOX", nil, MerchantFrame,
      "NigglesMerchantSearchTemplate");

    -- Create the masks for the Merchant Item buttons
    for buttonIdx = 1, MERCHANT_ITEMS_PER_PAGE do
      frame = CreateFrame("Frame", nil, _G["MerchantItem"..buttonIdx]);
      frame:SetFrameLevel(129);
      frame:SetAllPoints();
      texture = frame:CreateTexture(nil, "OVERLAY");
      texture:SetAllPoints(_G["MerchantItem"..buttonIdx.."ItemButton"]);
      texture:SetTexture(0, 0, 0, 0.8);
      texture = frame:CreateTexture(nil, "OVERLAY");
      texture:SetPoint("TOPLEFT",
        _G["MerchantItem"..buttonIdx.."NameFrame"], 1, -1);
      texture:SetWidth(105);
      texture:SetHeight(48);
      texture:SetTexture(0, 0, 0, 0.8);
      merchantItemMasks[buttonIdx] = frame;
    end
  end

  -- Check if Merchant Search is enabled
  if (NigglesDBC.settings.UiMerchantSearch) then
    -- Show the Merchant Search edit box
    merchantSearch:Show();

    -- Update the Merchant's frame, if shown
    if (MerchantFrame:IsShown()) then
      MerchantFrame_Update();
    end
  else
    -- Hide the Merchant Search edit box, if required
    if (merchantSearch ~= nil) then
      merchantSearch:Hide();
    end
  end

  return;
end

--
-- Function used to update the Merchant frame, based on the search text
--
function NigglesMerchantSearchUpdate()
  -- Local Variables
  local itemIdx;
  local itemMatch;
  local itemOffset = ((MerchantFrame.page-1)*MERCHANT_ITEMS_PER_PAGE);
  local searchText;
  local numItems = GetMerchantNumItems();
  local numCurrencies;
  local currencyLink;

  -- Check if searching is enabled
  if ((NigglesDBC.settings.UiMerchantSearch == true) and
      (merchantSearch ~= nil)) then
    -- Get the search text
    searchText = merchantSearch:GetText();
    if ((searchText == SEARCH) or (searchText == nil)) then
      searchText = "";
    else
      searchText = string.lower(searchText);
    end

    -- Set the vertex colour, depending on the buttons' text
    for buttonIdx = 1, MERCHANT_ITEMS_PER_PAGE do
      -- Initialise some variables
      itemIdx       = itemOffset+buttonIdx;
      itemMatch     = true;
      numCurrencies = GetMerchantItemCostInfo(itemIdx);

      -- Check if the name of the button's item matches the search string
      if ((searchText ~= "") and (itemIdx <= numItems)) then
        if (string.find(string.lower(GetMerchantItemInfo(itemIdx) or ""),
             searchText, 1, true) == nil) then
          itemMatch = false;
        end

        -- If the item's name doesn't match
        if ((not itemMatch) and (numCurrencies > 0)) then
          -- Check if any alternative currency item matches the search string
          for currencyIdx = 1, numCurrencies do
            _, _, currencyLink = GetMerchantItemCostItem(itemIdx,
              currencyIdx);
            if ((currencyLink ~= nil) and
                (string.find(string.lower(string.match(currencyLink,
                  "|h%[(.*)%]|h|r") or ""), searchText, 1, true) ~= nil)) then
              itemMatch = true;
              break;
            end
          end
        end
      end

      -- Show/Hide the merchant item's mask
      merchantItemMasks[buttonIdx]:SetShown(not itemMatch);
    end

    -- Show/Hide the Clear button
    merchantSearch.clearButton:SetShown(searchText ~= "");
  end

  return;
end
