-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to initialise the changes to the Quest panels
--
function NigglesQuestsInit()
  -- Local Variables
  local idx = 1;
  local texture;

  -- Create a texture for each Quest Item button to indicate which is
  -- most valuable
  while (_G["QuestInfoItem"..idx] ~= nil) do
    -- Create the texture
    if (_G["QuestInfoItem"..idx.."Valuable"] == nil) then
      texture = _G["QuestInfoItem"..idx]:CreateTexture(
        "QuestInfoItem"..idx.."Valuable", "BORDER");
      if (texture ~= nil) then
        texture:SetTexture("Interface\\Minimap\\TRACKING\\Auctioneer");
        texture:SetSize(16, 16);
        texture:ClearAllPoints();
        texture:SetPoint("BOTTOMRIGHT", -2, 2);
        texture:Hide();
      end

      -- Move the button's text to the Artwork layer
      _G["QuestInfoItem"..idx.."Name"]:SetDrawLayer("ARTWORK");
    end

    -- Move onto the next button
    idx = idx+1;
  end

  -- Hook the function that updates all the Quest Info Item buttons
  hooksecurefunc("QuestInfo_Display", NigglesQuestsItemInfoUpdate);

  return;
end

--
-- Function to update the Quest Item buttons to indicate which reward is most
-- valuable when sold to a merchant
--
function NigglesQuestsItemInfoUpdate()
  -- Local Variables
  local choiceLink;
  local choiceCount;
  local choiceValue = {};
  local idx;
  local maxValue = 0;
  local numChoices = 0;
  local missingInfo = false;

  -- Check at least one button is shown in the Quest Info Rewards frame
  if (_G["QuestInfoItem1"]:IsShown()) then
    -- Check if the most valuable quest reward(s) should be indicated
    if (NigglesDBC.settings.UiQuestRewards) then
      idx = 1;
      while (_G["QuestInfoItem"..idx] ~= nil) do
        -- Check the button is showing a quest reward choice
        button = _G["QuestInfoItem"..idx];
        if ((button:IsShown()) and
            (button.type       == "choice") and
            (button.objectType == "item")) then
          -- Get the link for the button's item
          if (QuestInfoFrame.questLog) then
            choiceLink = GetQuestLogItemLink("choice", button:GetID());
            _, _, choiceCount = GetQuestLogChoiceInfo(button:GetID());
          else
            choiceLink = GetQuestItemLink("choice", button:GetID());
            _, _, choiceCount = GetQuestItemInfo("choice", button:GetID());
          end
          if ((choiceLink ~= nil) and (choiceCount ~= nil)) then
            -- Get the item's value
            _, _, _, _, _, _, _, _, _, _, choiceValue[idx] =
              GetItemInfo(choiceLink);
            if (choiceValue[idx] ~= nil) then
              choiceValue[idx] = choiceValue[idx]*choiceCount;
              if (choiceValue[idx] > maxValue) then
                maxValue = choiceValue[idx];
              end
            else
              maxValue    = 0;
              missingInfo = true;
              break;
            end
          else
            maxValue    = 0;
            missingInfo = true;
            break;
          end
          numChoices = numChoices+1;
        end

        -- Move on to the next button
        idx = idx+1;
      end
    end

    -- Show/Hide the textures that indicate which items are the most valuable
    idx = 1;
    while (_G["QuestInfoItem"..idx] ~= nil) do
      button = _G["QuestInfoItem"..idx];
      if ((button:IsShown()) and
          (button.type == "choice") and
          (button.objectType == "item") and
          (numChoices > 1) and
          (maxValue > 0) and
          (choiceValue[idx] == maxValue)) then
        _G["QuestInfoItem"..idx.."Valuable"]:Show();
      else
        _G["QuestInfoItem"..idx.."Valuable"]:Hide();
      end
      idx = idx+1;
    end

    -- Check if the value of any items couldn't be obtained
    if ((NigglesDBC.settings.UiQuestRewards) and
        (missingInfo)) then
      NigglesTimerCreate("QuestItemValue", NigglesQuestsItemInfoUpdate,
        nil, nil, 0.25, false);
    end
  end

  return;
end
