-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to add a name to the list of raid assistants
--
local function raidAssistsAddName(frame, name)
  -- Local Variables
  local list = Niggles.updatedOptions.raidAssists;

  -- Check the name isn't already in the list
  if (not tContains(list, name)) then
    -- Add the name to the list
     tinsert(list, name);
     sort(list);

     -- Update the scroll frame
     frame.scrollFrame.update();
  end

  return;
end

--
-- Function to remove a name from the list of raid assistants
--
local function raidAssistsDeleteName(self, arg1, arg2, checked)
  -- Local Variables
  local list = Niggles.updatedOptions.raidAssists;

  -- Delete the name from the list
  tDeleteItem(list, list[arg1]);

  -- Update the scroll frame
  NigglesOptionsRaidsAssists.scrollFrame.update();

  return;
end

--
-- Function to initialise the drop down menu for the role/status frames
--
local function raidAssistsDropDownInit(self, level)
  -- Local Variables
  local info = self.info;

  -- Initialise some variables
  wipe(info);
  info.arg1 = self;

  -- Work out which level of menu needs initialising
  if (level == 1) then
    -- Add the invitee's name
    info.text         = Niggles.updatedOptions.raidAssists[self.index];
    info.isTitle      = 1;
    info.notCheckable = 1;
    UIDropDownMenu_AddButton(info, level)

    -- Clear the title flags
    info.isTitle      = nil;
    info.disabled     = nil;
    info.notCheckable = 1;
    info.isNotRadio   = 1;

    info.text = L["Delete"];
    info.func = raidAssistsDeleteName;
    info.arg1 = self.index;
    UIDropDownMenu_AddButton(info, level);
  end
  
  return;
end

--
-- Function to update the Raid Assistants scroll frame
--
local function raidAssistsScrollFrameUpdate()
  -- Local Variables
  local buttons;
  local numButtons;
  local numRaidAssists;
  local raidAssists = Niggles.updatedOptions.raidAssists;
  local scrollFrame = NigglesOptionsRaidsAssistsScrollFrame;
  local topOffset;

  -- Check the scroll frame has buttons
  if ((scrollFrame.buttons ~= nil) and
      (#scrollFrame.buttons > 0)) then
    -- Initialise some variables
    numButtons     = #scrollFrame.buttons;
    numRaidAssists = #raidAssists;
    topOffset      = HybridScrollFrame_GetOffset(scrollFrame);

    -- Update the scroll frame's buttons
    for idx = 1, numButtons do
      button = scrollFrame.buttons[idx];
      if (topOffset+idx <= numRaidAssists) then
        button:SetText(raidAssists[topOffset+idx]);
        button:SetID(idx);
        button:Show();
      else
        button:Hide();
      end
    end

    -- Update the scroll frame's height
    HybridScrollFrame_Update(scrollFrame,
      numRaidAssists*scrollFrame.buttons[1]:GetHeight(),
      numButtons*scrollFrame.buttons[1]:GetHeight());
  else
    HybridScrollFrame_Update(scrollFrame, 0, 0);
  end

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to process clicks on buttons in the 'Raid Assistants' list
--
function NigglesOptionsRaidsAssistsBtnOnClick(self, mouseButton)
  -- Local Variables
  local dropDown = self:GetParent():GetParent():GetParent().dropDown;

  -- Check which button was clicked
  if (mouseButton == "LeftButton" ) then
    HideDropDownMenu(1);
  elseif (mouseButton == "RightButton") then
    -- Show/Hide the drop down menu
    if ((dropDown:IsShown()) and (self:GetID() ~= dropDown.index)) then
      HideDropDownMenu(1);
    end
    dropDown.index       = self:GetID();
    dropDown.scrollFrame = self:GetParent();
    ToggleDropDownMenu(1, nil, dropDown, "cursor", 12, -1);
  end

  return;
end

--
-- Function called when a 'Raids Assistants' button is loaded
--
function NigglesOptionsRaidsAssistsBtnOnLoad(self)
  -- Register for left and right clicks
  self:RegisterForClicks("LeftButtonUp", "RightButtonUp");

  -- Clear the pushed offset
  self:SetPushedTextOffset(0, 0);
  
  return;
end

--
-- Function to perform custom initialisation for objects in the Options
-- Raids panel
--
function NigglesOptionsRaidsInit(object, setting)
  -- Initialise the Raid Assists frame
  NigglesCharacterListInit(object, L["RaidsAssists"],
    "NigglesOptionsRaidsAssistsBtnTemplate",
    raidAssistsScrollFrameUpdate, raidAssistsAddName,
    raidAssistsDropDownInit);

  return;
end

--
-- Function to promote raid assistants
--
function NigglesRaidRosterPromoteAssists()
  -- Local Variables
  local numMembers = GetNumGroupMembers();
  local name;
  local rank;

  -- Check if there are any raid assistants
  if (next(NigglesDB.raidAssists) ~= nil) then
    -- Process all the raid members
    for idx = 1, numMembers do
      name, rank = GetRaidRosterInfo(idx);
      if ((name ~= nil) and
          (rank == 0) and
          (NigglesTableContains(NigglesDB.raidAssists, name))) then
        PromoteToAssistant(name, true);
      end
    end
  end

  return;
end
